@@ -74,9 +74,14 @@ class AbCacheConfig:
     app_platform: str
     app_hash: str
 
-    ab_version: str = None  # Override AB version in url for ROW
     auth_credential: str = None  # JWT token for JP/EN, Base64 encoded JWT token for ROW
 
+    row_ab_version: str = None  # Override AB version in url for ROW
+
+    asset_host: str = None  # Override AssetBundle host for EN/JP
+    asset_version: str = None  # Override AssetBundle version for EN/JP
+    asset_hash: str = None  # Override AssetBundle hash for EN/JP
+
     version: Tuple[int, int, int] = (0, 0, 0)  # Internal versioning
 
     @property
@@ -84,10 +89,15 @@ class AbCacheConfig:
         if self.app_region in REGION_JP_EN:
             return self.auth_credential
         else:
-            return b64decode(self.auth_credential).decode()
+            try:
+                return b64decode(self.auth_credential).decode()
+            except:
+                return None
 
     @property
     def auth_jwt_payload(self) -> dict:
+        if not self.auth_jwt:
+            return {}
         header, payload, signature = self.auth_jwt.split(".")
         return json.loads(b64decode(payload + "=="))
 
@@ -97,7 +107,7 @@ class AbCacheConfig:
 
     @property
     def auth_available(self):
-        return self.auth_credential is not None
+        return self.auth_userID is not None
 
     @property
     def cache_version(self):
@@ -111,6 +121,15 @@ class AbCacheConfig:
     def is_up_to_date(self):
         return self.cache_version == __version_tuple__
 
+    @property
+    def need_client_header_update(self):
+        if self.app_region in REGION_JP_EN:
+            if self.app_region in {"jp"}:
+                return not (self.asset_hash and self.asset_host and self.asset_version)
+            else:
+                return not (self.asset_hash and self.asset_version)
+        return False
+
 
 @dataclass
 class AbCacheEntry(dict):
@@ -122,9 +141,10 @@ class AbCacheEntry(dict):
     crc: int
     fileSize: int
     dependencies: List[str]
-    paths: List[str]
     isBuiltin: bool
-    # TW, KR only
+    isRelocate: Optional[bool] = None  # Added in JP 5.4.0
+    paths: Optional[List[str]] = None  # Removed in ROW
+    # ROW only
     md5Hash: Optional[str] = None
     downloadPath: Optional[str] = None
 
@@ -132,7 +152,7 @@ class AbCacheEntry(dict):
 @dataclass
 class AbCacheIndex(dict):
     version: str
-    os: Optional[str] = None  # Undefined in KR, TW
+    os: Optional[str] = None  # Undefined in ROW
     bundles: Mapping[str, AbCacheEntry] = None
 
 
@@ -156,7 +176,7 @@ class SekaiAppVersion:
 @dataclass
 class SekaiSystemData:
     serverDate: int
-    # Undefined in KR, TW
+    # These are all undefined in ROW
     timezone: Optional[str] = None
     profile: Optional[str] = None
     maintenanceStatus: Optional[str] = None
@@ -221,6 +241,7 @@ class SekaiUserAuthData:
     obtainedBondsRewardIds: Optional[list] = None  # JP 4.0+
 
     # ROW
+    cdnVersion: Optional[int] = None
     configs: Optional[List[dict]] = None
 
 
@@ -255,6 +276,12 @@ class AbCache(Session):
     def SEKAI_API_ENDPOINT(self):
         match self.config.app_region:
             case "jp":
+                if self.database.sekai_gameversion_data:
+                    url = self.database.sekai_gameversion_data.domain
+                    url = url.replace(
+                        "{0}", self.database.sekai_gameversion_data.profile
+                    )
+                    return "https://%s" % url
                 return "https://production-game-api.sekai.colorfulpalette.org"
             case "en":
                 return "https://n-production-game-api.sekai-en.com"
@@ -263,7 +290,7 @@ class AbCache(Session):
             case "kr":
                 return "https://mkkorea-obt-prod01-cdn.bytedgame.com"
             case "cn":
-                return "https://mkcn-prod-public-30001-1.dailygn.com"
+                return "https://mkcn-prod-public-60001-1.dailygn.com"
             case _:
                 raise NotImplementedError
 
@@ -285,37 +312,67 @@ class AbCache(Session):
             case _:
                 raise NotImplementedError
 
+    @property
+    def SEKAI_AB_ROW_CDN(self):
+        match self.config.app_region:
+            case "tw":
+                return "https://lf16-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245"
+            case "kr":
+                return (
+                    "https://lf19-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248"
+                )
+            case "cn":
+                return "https://lf3-mkcncdn-tos.dailygn.com/obj/sf-game-lf/gdl_app_5236"
+            case _:
+                raise NotImplementedError
+
+    @property
+    def SEKAI_AB_ROW_PATH(self):
+        match self.config.app_region:
+            case "tw":
+                return "online"
+            case "kr":
+                return "kr_online"
+            case "cn":
+                return "cn_online"
+            case _:
+                raise NotImplementedError
+
+    @cached_property
+    def SEKAI_AB_ROW_VERSION_NUMBER(self):
+        url = f"{self.SEKAI_AB_ROW_CDN.replace('sf-game','rt-game')}"
+        if self.config.app_region in {"cn"}:
+            url = f"{url}/Mainland"
+        else:
+            url = f"{url}/Oversea"
+        url = f"{url}/{self.SEKAI_AB_ROW_VERSION}/Release/{self.SEKAI_AB_ROW_PATH}/{self.SEKAI_APP_PLATFORM}/version"
+        resp = self.request("GET", url)
+        resp.raise_for_status()
+        return int(resp.text)
+
     @property
     def SEKAI_AB_INFO_ENDPOINT(self):
         match self.config.app_region:
             case "jp":
-                return f"https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle-info.sekai.colorfulpalette.org/api/version/{self.SEKAI_ASSET_VERSION}/os/{self.config.app_platform}"
+                return f"https://{self.SEKAI_API_PROFILE}-{self.SEKAI_AB_HOST_HASH}-assetbundle-info.sekai.colorfulpalette.org/api/version/{self.SEKAI_ASSET_VERSION}/{self.SEKAI_AB_HASH}/os/{self.config.app_platform}"
             case "en":
                 return f"https://assetbundle-info.sekai-en.com/api/version/{self.SEKAI_ASSET_VERSION}/os/{self.config.app_platform}"
-            case "tw":  # NOTE: Android only
-                return f"https://lf16-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/online/android49/AssetBundleInfoNew.json"
-            case "kr":  # NOTE: Android only
-                return f"https://lf16-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/kr_online/android53/AssetBundleInfoNew.json"
-            case "cn":  # NOTE: Android only
-                return f"https://lf3-mkcncdn-tos.dailygn.com/obj/sf-game-lf/gdl_app_5236/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/cn_online/android50/AssetBundleInfoNew.json"
-            case _:
-                raise NotImplementedError
+        if self.config.app_region in REGION_ROW:
+            return f"{self.SEKAI_AB_ROW_CDN}/AssetBundle/{self.SEKAI_AB_ROW_VERSION}/Release/{self.SEKAI_AB_ROW_PATH}/{self.SEKAI_APP_PLATFORM}{self.SEKAI_AB_ROW_VERSION_NUMBER}/AssetBundleInfoNew.json"
+        else:
+            raise NotImplementedError
 
     @property
     def SEKAI_AB_ENDPOINT(self):
         match self.config.app_region:
             case "jp":
-                return f"https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle.sekai.colorfulpalette.org/"
+                return f"https://{self.SEKAI_API_PROFILE}-{self.SEKAI_AB_HOST_HASH}-assetbundle.sekai.colorfulpalette.org/"
             case "en":
                 return f"https://assetbundle.sekai-en.com/"
-            case "tw":  # NOTE: Android only
-                return f"https://lf16-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/online/android1/"
-            case "kr":  # NOTE: Android only
-                return f"https://lf16-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/kr_online/android1/"
-            case "cn":  # NOTE: Android only
-                return f"https://lf3-j1gamecdn-cn.dailygn.com/obj/sf-game-lf/gdl_app_5236/AssetBundle/{self.config.ab_version or self.config.app_version}/Release/cn_online/android1/"
-            case _:
-                raise NotImplementedError
+        if self.config.app_region in REGION_ROW:
+            return f"{self.SEKAI_AB_ROW_CDN}/AssetBundle/{self.SEKAI_AB_ROW_VERSION}/Release/{self.SEKAI_AB_ROW_PATH}/"
+        else:
+            raise NotImplementedError
 
     def _update_request_headers(self):
         self.headers.update(
@@ -345,6 +402,10 @@ class AbCache(Session):
     def SEKAI_APP_VERSION(self):
         return self.config.app_version
 
+    @property
+    def SEKAI_AB_ROW_VERSION(self):
+        return self.config.row_ab_version or self.SEKAI_APP_VERSION
+
     @property
     def SEKAI_APP_PLATFORM(self):
         return self.config.app_platform
@@ -355,17 +416,29 @@ class AbCache(Session):
 
     @property
     def SEKAI_ASSET_VERSION(self):
-        return self.database.sekai_system_data.app_version_dict[
-            self.SEKAI_APP_VERSION
-        ].assetVersion
+        return (
+            self.config.asset_version
+            or self.database.sekai_system_data.app_version_dict[
+                self.SEKAI_APP_VERSION
+            ].assetVersion
+        )
 
     @property
     def SEKAI_AB_HASH(self):
-        return self.database.sekai_user_auth_data.assetHash
+        return self.config.asset_hash or self.database.sekai_user_auth_data.assetHash
 
     @property
     def SEKAI_AB_HOST_HASH(self):
-        return self.database.sekai_gameversion_data.assetbundleHostHash
+        return (
+            self.config.asset_host
+            or self.database.sekai_gameversion_data.assetbundleHostHash
+        )
+
+    @property
+    def SEKAI_API_PROFILE(self):
+        if self.database.sekai_gameversion_data:
+            return self.database.sekai_gameversion_data.profile
+        return "production"
 
     @property
     def SEKAI_API_SYSTEM_DATA(self):
@@ -429,21 +502,11 @@ class AbCache(Session):
                     f"{self.SEKAI_API_ENDPOINT}/api/{path}"
                     for path in self.database.sekai_user_auth_data.suiteMasterSplitPath
                 ]
-            # NOTE: All ROW servers have hardcoded master data URLs in *one* file as of the time of writing.
-            # NOTE: Some of the MessagePack schemas is also *omitted* in these endpoints. Good luck parsing them.
-            case "tw":
-                return [
-                    "https://lf21-mkovscdn-sg.bytedgame.com/obj/sf-game-alisg/gdl_app_5245/MasterData/60001/master-data-138.info"
-                ]
-            case "kr":
-                return [
-                    "https://lf19-mkkr.bytedgame.com/obj/sf-game-alisg/gdl_app_292248/MasterData/60001/master-data-158.info"
-                ]
-            case "cn":
-                # XXX: Unverified. Game's still unreleased in CN
-                return [
-                    "https://lf3-j1gamecdn-cn.dailygn.com/obj/sf-game-lf/gdl_app_5236/MasterData/60001/master-data-138.info"
-                ]
+        if self.config.app_region in REGION_ROW:
+            self.raise_for_auth()
+            return f"{self.SEKAI_AB_ROW_CDN}/{self.SEKAI_AB_ROW_PATH}/master-data-{self.database.sekai_user_auth_data.cdnVersion}.info"
+        else:
+            raise NotImplementedError
 
     @property
     def SEKAI_API_INFORMATION(self):
@@ -484,6 +547,9 @@ class AbCache(Session):
                 logger.error("response=%s" % content)
             except:
                 logger.error("response=%s" % resp.content)
+            logger.error(
+                "Unexpected server-side error. Refer to https://github.com/mos9527/sssekai/wiki#debugging-abcache for more information."
+            )
             resp.raise_for_status()
         return resp
 
@@ -527,10 +593,9 @@ class AbCache(Session):
             if self.config.app_region in REGION_ROW:
                 self.database.sekai_user_data = fromdict(SekaiUserData, data, False)
         else:
-            if self.config.app_region in REGION_JP_EN:
-                raise AbCacheUserNotAuthenticatedError(
-                    "Auth config is required for accessing JP/EN game server content"
-                )
+            logger.warning(
+                "No valid auth credential provided. User auth data will not be updated."
+            )
 
     def _update_system_data(self):
         logger.debug("Updating system data")
@@ -552,25 +617,6 @@ class AbCache(Session):
             data = self.response_to_dict(resp)
             self.database.sekai_gameversion_data = fromdict(SekaiGameVersionData, data)
 
-    def _update_abcache_index(self) -> AbCacheIndex:
-        logger.debug("Updating Assetbundle index")
-        resp = self.request_packed("GET", self.SEKAI_AB_INFO_ENDPOINT)
-        data = self.response_to_dict(resp)
-        self.database.sekai_abcache_index = fromdict(AbCacheIndex, data)
-        return self.database.sekai_abcache_index
-
-    def _update_signatures(self):
-        """Only required for JP for the initial setup. After which the game would cache the signature/cookies."""
-        if self.config.app_region in {"jp"}:
-            logger.debug("Updating signatures")
-            # XXX: Clear all cookies, it 403s otherwise. Does this trigger a rate limit/expiry check since our session would be tracked?
-            self.headers["Cookie"] = ""
-            self.cookies.clear()
-            resp = self.request_packed("POST", self.SEKAI_ISSUE_SIGNATURE_ENDPOINT)
-            self.headers["Cookie"] = resp.headers["Set-Cookie"]
-            # HACK: Per RFC6265, Cookies should not be visible to subdomains since it's not set with Domain attribute (https://github.com/psf/requests/issues/2576)
-            # But the other endpoints uses it nontheless. So we have to set it manually.
-
     def __init__(self, config: Optional[AbCacheConfig] = None):
         super().__init__()
         self.database = SSSekaiDatabase()
@@ -580,28 +626,21 @@ class AbCache(Session):
         self.config.version = __version_tuple__
 
     def update_client_headers(self):
-        """Authenticate the user and update client headers WITHOUT updating the AssetBundle Index.
-
-        NOTE:
-            - For JP/EN servers, `auth_credential` NEED to be set or otherwise you won't be able to do anything.
-            - For ROW servers, it's OK to leave them empty.
-                - To access user-level data, however, you still NEED to be authenticated.
-        """
+        """Authenticate the user and update client headers."""
         logger.debug("Updating metadata")
         logger.debug("Set config: %s" % self.config)
         try:
-            self._update_signatures()
+            self.update_signatures()
+            self._update_gameversion_data()
             self._update_system_data()
         except HTTPError as e:
-            logger.warning("Attempting to update signatures: %s" % e)
-            self._update_signatures()
-            self._update_system_data()
+            logger.error("Signature update failure: %s" % e)
+            raise e
         if self.config.app_region in REGION_JP_EN:
             version_newest = self.database.sekai_system_data.appVersions[-1]
             logger.debug("Newest App version: %s" % version_newest)
             if version_newest.appVersion != self.SEKAI_APP_VERSION:
                 logger.warning("App version mismatch. This may cause issues.")
-        self._update_gameversion_data()
         self._update_user_auth_data()
         if self.is_authenticated:
             self.headers.update(
@@ -613,6 +652,31 @@ class AbCache(Session):
             )
         return self.headers
 
+    def update_abcache_index(self) -> AbCacheIndex:
+        """Update the AssetBundle Index."""
+        logger.debug("Updating Assetbundle index")
+        if self.config.app_region in REGION_JP_EN:
+            logger.info("Sekai AssetBundle version: %s" % self.SEKAI_ASSET_VERSION)
+            if self.config.app_region in {"jp"}:
+                logger.info("Sekai AssetBundle host hash: %s" % self.SEKAI_AB_HOST_HASH)
+        resp = self.request_packed("GET", self.SEKAI_AB_INFO_ENDPOINT)
+        data = self.response_to_dict(resp)
+        self.database.sekai_abcache_index = fromdict(AbCacheIndex, data)
+        return self.database.sekai_abcache_index
+
+    def update_signatures(self):
+        """Update client signatures. Only required for JP for the initial setup.
+        After which the game would cache the signature/cookies."""
+        if self.config.app_region in {"jp"}:
+            logger.debug("Updating signatures")
+            # XXX: Clear all cookies, it 403s otherwise. Does this trigger a rate limit/expiry check since our session would be tracked?
+            self.headers["Cookie"] = ""
+            self.cookies.clear()
+            resp = self.request_packed("POST", self.SEKAI_ISSUE_SIGNATURE_ENDPOINT)
+            self.headers["Cookie"] = resp.headers["Set-Cookie"]
+            # HACK: Per RFC6265, Cookies should not be visible to subdomains since it's not set with Domain attribute (https://github.com/psf/requests/issues/2576)
+            # But the other endpoints uses it nontheless. So we have to set it manually.
+
     def update(self):
         """Update the cache with the latest AssetBundle Index data from the server.
 
@@ -621,19 +685,16 @@ class AbCache(Session):
             Please refer to `update_client_headers` for more information.
         """
         self.update_client_headers()
-        self._update_abcache_index()
-        if self.config.app_region in REGION_JP_EN:
-            logger.debug("Sekai AssetBundle version: %s" % self.SEKAI_ASSET_VERSION)
-            logger.debug("Sekai AssetBundle host hash: %s" % self.SEKAI_AB_HOST_HASH)
+        self.update_abcache_index()
 
     def save(self, f: BinaryIO):
         self._update_request_headers()
         self.database.cached_headers = self.headers
-        logger.debug("Saving cache")
+        logger.info("Saving cache: %s" % self)
         dump(self.database, f)
 
     def load(self, f: BinaryIO):
-        logger.debug("Loading cache")
+        logger.info("Loading cache")
         self.database = load(f)
         if not self.database.config.is_up_to_date:
             logger.warning(
@@ -641,7 +702,7 @@ class AbCache(Session):
                 % (self.database.config.cache_version_string, __version__)
             )
         self.headers.update(self.database.cached_headers)
-        logger.debug("Cache loaded: %s" % self)
+        logger.info("Cache loaded: %s" % self)
 
     @staticmethod
     def from_file(f: BinaryIO):
@@ -650,9 +711,7 @@ class AbCache(Session):
         return cache
 
     def __repr__(self) -> str:
-        return (
-            f"AbCache(config={self.config} bundles={len(self.abcache_index.bundles)})"
-        )
+        return f"AbCache(config={self.config} bundles={len(self.abcache_index.bundles) if self.abcache_index is not None else 'N/A'})"
 
     def get_entry_by_bundle_name(self, bundleName: str) -> AbCacheEntry:
         return self.abcache_index.bundles.get(bundleName, None)
@@ -661,7 +720,8 @@ class AbCache(Session):
         if self.config.app_region in REGION_JP_EN:
             return self.SEKAI_AB_ENDPOINT + self.SEKAI_AB_BASE_PATH + entry.bundleName
         else:
-            return self.SEKAI_AB_ENDPOINT + entry.bundleName
+            # https://github.com/mos9527/sssekai/issues/28
+            return self.SEKAI_AB_ENDPOINT + entry.downloadPath + "/" + entry.bundleName
 
     def get_or_update_dependency_tree_flatten(self, bundleName: str, deps: set = None):
         """Get a flattened set of asset dependency bundle names (including itself) for a given entry.
