@@ -1,36 +1,34 @@
 from sssekai.unity.AssetBundle import load_assetbundle
-import os, json
+import os, json, tqdm
 
 
 def main_mvdata(args):
     from UnityPy.enums import ClassIDType
 
-    source = args.infile
+    source = args.indir
     source = os.path.expanduser(source)
     source = os.path.abspath(source)
+    outfile = os.path.abspath(args.outfile)
     os.chdir(source)
     mvdata_keys = sorted(os.listdir(source))
     mvdata_items = list()
-    for key in mvdata_keys:
+    for key in tqdm.tqdm(mvdata_keys):
         try:
             with open(key, "rb") as f:
                 env = load_assetbundle(f)
                 for obj in env.objects:
                     if obj.type == ClassIDType.MonoBehaviour:
                         data = obj.read()
-                        if data.m_Name == "data":
-                            typetree = obj.read_typetree()
-                            typetree = {
-                                k: v
-                                for k, v in typetree.items()
-                                if not k.startswith("m_")
-                            }
+                        typetree = obj.read_typetree()
+                        typetree = {
+                            k: v for k, v in typetree.items() if not k.startswith("m_")
+                        }
+                        if "name" in typetree:
                             mvdata_items.append(typetree)
-                            break
         except Exception as e:
             print(f"skipping {key}: {e}")
-    outdir = os.path.dirname(args.outdir)
+    outdir = os.path.dirname(outfile)
     if outdir:
         os.makedirs(outdir, exist_ok=True)
-    with open(args.outdir, "w", encoding="utf-8") as f:
+    with open(outfile, "w", encoding="utf-8") as f:
         json.dump(mvdata_items, f, indent=4, ensure_ascii=False)
