@@ -8,11 +8,16 @@ from UnityPy.classes import (
     AnimationClip,
     StreamedClip,
     GenericBinding,
+    # ---
+    Vector3Curve,
+    QuaternionCurve,
+    FloatCurve,
 )
 from UnityPy.classes.math import Vector3f as Vector3, Quaternionf as Quaternion
 from UnityPy.streams.EndianBinaryReader import EndianBinaryReader
 from logging import getLogger
 from enum import IntEnum
+from zlib import crc32
 
 logger = getLogger(__name__)
 
@@ -313,7 +318,8 @@ def _read_clip_keyframe(
             result.outSlope = float("inf")
             result.inSlope = float("inf")
         elif binding.isPPtrCurve:
-            raise NotImplementedError("PPtr curves not yet implemented")
+            # !! TODO PPtr curves not yet implemented
+            pass
         else:
             # Default to float curves
             key = next(keys)
@@ -342,6 +348,26 @@ class CurveHelper:
         rhs = lhs.next
         return KeyframeHelper.interpolate(t, lhs, rhs)
 
+    def resample_dense(self, times: List[float]) -> "CurveHelper": # O(n log n). Can be done linearly though.
+        """Resamples the curve at given times **as a Dense curve**, losing all tangent information.
+
+        This guarantees accuracy when played back at a constant sample rate at the cost of space
+        and interpolation information.
+
+        Args:
+            times (List[float]): Times to sample the curve at
+
+        Returns:
+            CurveHelper: A new CurveHelper with the sampled values
+        """
+        data = [self.evaluate(t) for t in times]
+        slope = from_floats(*([0] * num_floats(data[0])))
+        data = [KeyframeHelper(t, self.Binding.typeID, value, isDense=True, inSlope=slope, outSlope=slope) for t, value in zip(times, data)]
+        for i in range(1, len(data)):
+            data[i].prev = data[i - 1]
+            data[i - 1].next = data[i]
+        return CurveHelper(self.Binding, data)
+
     @property
     def Duration(self):
         return self.Data[-1].time - self.Data[0].time
@@ -376,13 +402,15 @@ class AnimationHelper:
     # Dict[internal hash, Curve]
     RawCurves: Dict[int, CurveHelper] = field(default_factory=dict)
     # Curves[Attribute][Path Hash] = Curve
-    Curves: Dict[int, Dict[float, CurveHelper]] = field(
+    Curves: Dict[int, Dict[int, CurveHelper]] = field(
         default_factory=lambda: defaultdict(dict)
     )
     # CurvesT[Path Hash][Attribute] = Curve
-    CurvesT: Dict[int, Dict[float, CurveHelper]] = field(
+    CurvesT: Dict[int, Dict[int, CurveHelper]] = field(
         default_factory=lambda: defaultdict(dict)
     )
+    # From path id to the full path
+    InvCRC: Dict[int, str] = field(default_factory=dict)
 
     @staticmethod
     def hash_of(binding: GenericBinding):
@@ -392,7 +420,7 @@ class AnimationHelper:
         hs = self.hash_of(binding)
         if not hs in self.RawCurves:
             curve = CurveHelper(binding)
-            # i love refcounting
+            # Refcounted
             self.RawCurves[hs] = curve
             self.Curves[binding.attribute][binding.path] = curve
             self.CurvesT[binding.path][binding.attribute] = curve
@@ -429,6 +457,62 @@ class AnimationHelper:
     def PPtrCurves(self):
         raise NotImplementedError
 
+    def process_legacy(self, src: AnimationClip) -> "AnimationHelper":
+        """Reads pre-build AnimationClip data and converts it to an Animation
+
+        NOTE: This only works with Legacy AnimationClips, where the animation data
+        is not compacted into a MuscleClip and is instead stored in the AnimationClip itself.
+
+        Note:
+            `Animation` in this context is NOT a part of Unity (which coincidentally also has an `Animation` class)
+            Read `sssekai.unity.Animation`'s docs for more information
+
+        Args:
+            src (AnimationClip): An Unity AnimationClip object
+
+        Returns:
+            sssekai.unity.Animation
+        """
+
+        def process_curves(
+            curves: List[Vector3Curve | QuaternionCurve | FloatCurve], attribute: int
+        ):
+            for curve in curves:
+                binding = GenericBinding(
+                    attribute or curve.attribute,
+                    None,
+                    False,
+                    crc32(curve.path.encode("utf-8")),
+                    None,
+                )
+                binding.typeID = ClassIDType.Transform
+                self.InvCRC[binding.path] = curve.path
+                keyframes = [
+                    KeyframeHelper(
+                        time=key.time,
+                        typeID=binding.typeID,
+                        value=key.value,
+                        inSlope=key.inSlope,
+                        outSlope=key.outSlope,
+                    )
+                    for key in curve.curve.m_Curve
+                ]
+                self.get_curve(binding).Data = keyframes
+            pass
+
+        process_curves(src.m_PositionCurves, kBindTransformPosition)
+        process_curves(src.m_RotationCurves, kBindTransformRotation)
+        process_curves(src.m_ScaleCurves, kBindTransformScale)
+        process_curves(src.m_EulerCurves, kBindTransformEuler)
+        process_curves(src.m_FloatCurves, None)
+        # ! TODO: PPtr curves not yet supported
+        # process_curves(src.m_PPtrCurves, ClassIDType.PPtr)
+        for curve in self.RawCurves.values():
+            for i in range(1, len(curve.Data)):
+                curve.Data[i].prev = curve.Data[i - 1]
+                curve.Data[i - 1].next = curve.Data[i]
+        return self
+
     def process(self, src: AnimationClip) -> "AnimationHelper":
         """Reads post-build AnimationClip data and converts it to an Animation
 
@@ -558,7 +642,10 @@ class AnimationHelper:
         helper = AnimationHelper(
             src.m_Name, src.m_MuscleClip.m_StopTime, src.m_SampleRate
         )
-        helper.process(src)
+        if src.m_Legacy:
+            helper.process_legacy(src)
+        else:
+            helper.process(src)
         return helper
 
 
